/***************************************************************************
                          base64.c  -  description
                             -------------------
    begin                : Sat Oct 11 2003
    copyright            : (C) 2003 by Eric W. Sarjeant
    email                : eric@sarjeant.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "base64.h"

// load alphabet of b64 characters
#define BASE64_LEN 64
#define BASE64_PAD '='
char b64alp[BASE64_LEN] = {'A', 'B', 'C', 'D', 'E', 'F', 'G', 'H',
                           'I', 'J', 'K', 'L', 'M', 'N', 'O', 'P',
                           'Q', 'R', 'S', 'T', 'U', 'V', 'W', 'X',
                           'Y', 'Z', 'a', 'b', 'c', 'd', 'e', 'f',
                           'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
                           'o', 'p', 'q', 'r', 's', 't', 'u', 'v',
                           'w', 'x', 'y', 'z', '0', '1', '2', '3',
                           '4', '5', '6', '7', '8', '9', '+', '/'};

#define BASE64_SEGSIZE 3    // size of segment for processing BASE64
#define BASE64_BLOCKSIZE 4  // size of encoded B64 data
#define BASE64_LINESIZE 76  // number of chars per line

// lookup a character in the BASE64 map
short get_base64_value(char c);
    
/**
 * Accept a pointer to an array of data and return that data in BASE64
 * format. Note that the length of data to encode is required; it is
 * possible to have NULL elements in the encoded data.
 *
 * INPUT
 *   char *inp   Original data to encode.
 *   int  inplen Length of data to select from *inp for encoding
 *
 * OUTPUT
 *   char *      Original data in BASE64 format.
 **/
extern char *encodeBASE64(char *inp, int inplen) {

    char c = '\0';            // single char read for encoding
    char *blkchr = NULL;      // BASE64 chr block
    int blklen = 0;           // current length of BASE64 block
    long blkraw = 0;          // raw concat of binary block data
    unsigned short blkb4 = 0; // 6-bit segments poked from blkraw
    int enclen = 0;           // total encoding length
    int inppos = 0;           // position in input string
    int encpos = 0;           // actual position in output string
    int linepos = 0;          // current position relative to line
    char *encstr = NULL;      // final output of B64 data

    // guess encoding length and malloc
    enclen = (((inplen / BASE64_SEGSIZE) * BASE64_BLOCKSIZE) + (inplen % BASE64_SEGSIZE > 0 ? BASE64_BLOCKSIZE : 0));
    enclen += (enclen % BASE64_LINESIZE) + 1;  // NL chars        
    encstr = (char *)calloc(enclen, sizeof(char));

    // setup block for B64 data
    blkchr = (char *)calloc(BASE64_SEGSIZE, sizeof(char));
    
    // primary encode loop
    while((c = (char)inp[inppos++]) != '\0') {

        blkchr[blklen++] = c;

        // look for a full block or NULL data terminator
        if ((BASE64_SEGSIZE == blklen) | (inp[inppos] == '\0')) {

            // newline
            if (linepos == BASE64_LINESIZE) {
                encstr[encpos++] = '\n';
                linepos = 0;
            }
            
            // raw binary value of concat blocks
            blkraw = ((blkchr[0] << 16) | (blkchr[1] << 8) | blkchr[2]);

            // shift blocks into encstr
            blkb4 = (blkraw >> 18| 0xFFC0) ^ 0xFFC0;
            encstr[encpos + 0] = b64alp[blkb4];

            blkb4 = ((blkraw >> 12) | 0xFFC0) ^ 0xFFC0;
            encstr[encpos + 1] = b64alp[blkb4];
            
            blkb4 = ((blkraw >> 6) | 0xFFC0) ^ 0xFFC0;
            encstr[encpos + 2] = b64alp[blkb4];

            blkb4 = ((blkraw) | 0xFFC0) ^ 0xFFC0;
            encstr[encpos + 3] = b64alp[blkb4];

            // padding?
            if (blkchr[2] == 0x00) {
                encstr[encpos + 3] = BASE64_PAD;
            }

            if (blkchr[1] == 0x00 && blkchr[2] == 0x00) {
                encstr[encpos + 2] = BASE64_PAD;
            }
                            
            // move into the next block
            encpos += BASE64_BLOCKSIZE;            
            linepos += BASE64_BLOCKSIZE;

            blkchr = (char *)memset(blkchr, 0x00, BASE64_SEGSIZE);
            blklen = 0;
            
        }
        
    }

    free(blkchr);

    return encstr;
    
}

/**
 * Reverse lookup of a character in the BASE64 map; return the location
 * of that character in the map. If character cannot be found, then expect
 * -1 here.
 *
 * INPUT
 *  char  c   Character to query the map for.
 *
 * OUTPUT
 *  short     Location in the map or -1 if location not found.
 **/
short get_base64_value(char c) {

    short v = -1;
    short l = 0;

    for (l=0; l < BASE64_LEN; l++) {
        if (b64alp[l] == c) {
            v = l;
        }
    }

    return v;
    
}

/**
 * Accept a NULL terminated pointer of char and return the BASE64 decoding
 * of that value. Very little sanity check takes place here, so if the
 * original data is misbehaved (eg: not evenly divisible by the BASE64
 * segment-size) then weird data results.
 *
 * INPUT:
 *     char *inp   Pointer of char to decode (NULL terminated).
 *     char **outp Pointer to pointer of char for output; pass by-ref
 *                 to contain BASE64 decoded data (ie: raw ASCII/binary).
 *                 Remember; this is by-ref so you must free() this
 *                 pointer when you are done with it.
 *
 * OUTPUT:
 *     int         Length of decoded data; data will also be NULL terminated
 *                 if caller is certain this is text-only. A value of -1
 *                 indicates something went wrong here; probably the process
 *                 could not allocate enough memory.
 **/
extern int decodeBASE64(char *inp, char **outp) {

    char c = '\0';              // chr from input
    int decpos = 0;             // position in string for decoding
    int decblk = 0;             // current blocksize of decode
    unsigned long decbit = 0;   // bit-shifted values for decode    
    int outpos = 0;             // position in output data
    char *outdata = NULL;       // output data
    
    // allocate output - a little oversized here...
    outdata = (char *)calloc(strlen(inp), sizeof(char));

    if (NULL == outdata) {
        return -1;
    }

    // init block offset inside incoming BASE64 data    
    decblk = BASE64_BLOCKSIZE;
    
    while (((c = (char)inp[decpos++]) != '\0')) {

        if (get_base64_value(c) >= 0) {
            decbit |= (get_base64_value(c) << (--decblk * 6));
        }

        // check for full block & decode
        if (0 == decblk) {

            outdata[outpos++] = (char)(decbit >> 16);
            outdata[outpos++] = (char)(decbit >> 8);
            outdata[outpos++] = (char)(decbit);
            
            decbit = 0;
            decblk = BASE64_BLOCKSIZE;
            
        }
        
    }                                              

    // check for remaining block
    if (decbit > 0) {
        outdata[outpos++] = (char)(decbit >> 16);
        outdata[outpos++] = (char)(decbit >> 8);
        outdata[outpos++] = (char)(decbit);
    }

    // terminate
    outdata[outpos++] = '\0';

    // by-ref
    *outp = *(&outdata);

    return outpos;
    
}
