/***************************************************************************
                          main.c  -  description
                             -------------------
    begin                : Sat Oct 11 10:52:57 EDT 2003
    copyright            : (C) 2003 by Eric W. Sarjeant
    email                : eric@sarjeant.com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <time.h>
#include "base64.h"

#define FILE_ENCODE "-encode"
#define FILE_DECODE "-decode"

#define MODE_ENCODE 1
#define MODE_DECODE 2

// number of blocks to read during file IO
#define IO_BLOCKSIZE 64

/**
 * Read the specified file into a char pointer data. String is always
 * NULL terminated. Expansion of the input buffer correlates to the
 * IO_BLOCKSIZE. Note that it must be possible to read the entire file
 * into memory at this point, if this fails a NULL is also returned
 * with feedback written to console.
 *
 *
 * INPUT
 *     char *fname   Filename to open
 *     char **data   Char pointer to fill with data
 *
 * OUTPUT
 *     char &data    By-ref value is stored in data or NULL on error
 **/
void read_file(char *fname, char **data) {

    FILE *fp = NULL;
    char *datablk = NULL;
    int flen = 0;

    if ((fp = fopen(fname, "r")) == NULL) {
        printf("IO Error: File \"%s\" could not be opened.\n", fname);
        return;
    } else {
        
        datablk = calloc(IO_BLOCKSIZE, sizeof(char));

        // obtain total amount of data
        if (fseek(fp, 0, SEEK_END) == 0) {

            flen = ftell(fp);

            // check memory
            if ((NULL == datablk) | ((*data = calloc(flen, sizeof(char))) == NULL)) {

                printf("Memory Allocation Error: Out of memory in read_file()\n");
                
                fclose(fp);
                free(datablk);
                
                return;
                
            }
            
            rewind(fp);
        }

        // read data
        while (fread(datablk, sizeof(char), IO_BLOCKSIZE, fp) > 0) {
            strncat(*data, datablk, IO_BLOCKSIZE);
            bzero(datablk, IO_BLOCKSIZE);
        }
        
    }
    
    fclose(fp);
    free(datablk);

}

void print_usage() {

    printf ("Usage: base64 [OPTION] [FILE]...\n\n");
    printf ("Encode/decode files in BASE64 format to standard out.\n\n");
    printf ("  -encode        Encode the file(s) to BASE64\n");
    printf ("  -decode        Decode the file(s) from BASE64\n");
    printf ("  -?, -h, -help  Display this help text\n\n");
    
}

/**
 * Primary command-line processing. Loop through the program parameters
 * and look for:
 *
 *    -encode   Switch to encode mode
 *    -decode   Switch to decode mode
 *    (other)   Treat as a file
 *
 * If (other) is specified and cannot be found, a textual error appears on
 * stdout.
 *
 *
 * INPUT
 *    int argc      Total number of command-line argument(s)
 *    char *argv[]  Values of the command-line argument(s)
 *
 * OUTPUT
 *    (none)
 **/
int main(int argc, char *argv[]) {
    
    char *fname = NULL;
    char *data = NULL;
    char *inp = NULL;
    int outlen = 0;
    int outloc = 0;
    int argl = 0;
    short mode = 0;

    fname = calloc(255, sizeof(char));

    for (argl = 1; argl < argc; argl++) {
        
        if (strcmp(argv[argl], FILE_ENCODE) == 0) {
            mode = MODE_ENCODE;
        } else if (strcmp(argv[argl], FILE_DECODE) == 0) {
            mode = MODE_DECODE;
        } else {

            strncpy(fname, argv[argl], 255);

            // pick the mode (encode|decode|err)
            switch (mode) {

                case MODE_ENCODE:

                    // read file
                    read_file(fname, &data);

                    if (NULL != data) {

                        inp = encodeBASE64(data, strlen(data));
                        
                        printf("---begin-encode-%s---\n%s\n---end-encode-%s---\n", fname, inp, fname);  // output
                        
                        free(inp);
                        free(data);
                        
                    }
                    
                    break;
                case MODE_DECODE:
                     
                     // read file
                     read_file(fname, &data);

                     if (NULL != data) {

                         if ((outlen = decodeBASE64(data, &inp)) > 0) {

                             printf("---begin-decode-%s---\n", fname);

                             for (outloc = 0; outloc < outlen; outloc++) {
                                 printf("%c", inp[outloc]);
                             }
                         
                             printf("\n---end-decode-%s---\n", fname);
                             
                         free(inp);
                         free(data);

                         } else {
                            printf("Error: Unable to decode file \"%s\"", fname);
                         }

                     }
                     
                     break;
                default:
                     print_usage();
                     
            }

        }
        
    }

    free(fname);

    return EXIT_SUCCESS;
    
}
